﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using Android.App;
using Android.Content;
using Android.Net;
using Android.Net.Wifi;
using Android.OS;
using Android.Runtime;
using Android.Views;
using Android.Widget;
using Newtonsoft.Json;
using NoxService.NWS;
using WalletLibrary;

namespace BitfiWallet
{
  public class NoxTManager : AsyncTickProcessor
  {
    public static Device noxDevice;

    Activity _activity; Context _appCtx;

    public bool WsRunning = false;

    public bool _BroadCastAction = true;

    public NoxTManager(Device _noxDevice) : base(_noxDevice)
    {
      noxDevice = _noxDevice;
      Sclear.Initiate();
    }
    public void Resume(Activity activity)
    {
      _activity = activity;
      _appCtx = _activity.ApplicationContext;
      _BroadCastAction = true;

      Task.Run(() =>
      {
        var started = Start();

        if (!started)
        {
          WsRunning = true;
        }

      });
    }

    public void Pause()
    {
      _BroadCastAction = false;

      Task.Run(() =>
      {
        var stopped = Stop();

        WsRunning = false;

      });
    }

    public void Addtime(double time_seconds = 30)
    {
      AddDelay(time_seconds);
    }

    protected override Func<Task<WorkResponse>> _ConnectionTicks()
    {
      return (async () =>
      {

        await Task.Delay(50); 

        bool networking = true;

        using (WifiManager wifi = (WifiManager)_activity.GetSystemService(Context.WifiService))
        using (ConnectivityManager cm = (ConnectivityManager)_activity.GetSystemService(Context.ConnectivityService))
        {
          if (wifi.IsWifiEnabled != true) wifi.SetWifiEnabled(true);

          while (networking && !cancellation.IsCancellationRequested)
          {
            using (NetworkInfo activeNetwork = cm.ActiveNetworkInfo)
            {
              if (activeNetwork != null && activeNetwork.IsConnected)
              {
                networking = false;
                var resp = new WorkResponse() { ValidationRate = "0", Action = ResponseActionType.SHOW_RATE };
                HandleWorkResponse(resp);
                return resp;
              }
            }

            using (WifiInfo wifiInfo = wifi.ConnectionInfo)
            using (NetworkInfo.DetailedState state = WifiInfo.GetDetailedStateOf(wifiInfo.SupplicantState))
            {

              if (state == NetworkInfo.DetailedState.Connecting) WebStatusNotification("now connecting...", false);

              if (state == NetworkInfo.DetailedState.Authenticating) WebStatusNotification("authenticating network...", false);

              if (state == NetworkInfo.DetailedState.ObtainingIpaddr) WebStatusNotification("obtaining ip address...", false);

              if (state == NetworkInfo.DetailedState.Disconnected)
              {
                networking = false;
                var resp = new WorkResponse() { ValidationRate = "There's no wifi connection", Action = ResponseActionType.NOWIFI };
                HandleWorkResponse(resp);

                Pause();
                return resp;
              }
            }


          }
        }

        return null;

      });
    }
    protected override void HandleWorkResponse(WorkResponse data)
    {
      try
      {

        base.HandleWorkResponse(data);

        if (!_BroadCastAction) return;

        switch (data.Action)
        {
          case ResponseActionType.CLOSE:
            TimoutNotification();
            break;
          case ResponseActionType.NONE:
            break;
          case ResponseActionType.NOWIFI:
            WebStatusNotification(data.ValidationRate, true);
            WsRunning = false;
            break;
          case ResponseActionType.REQUEST:
            TaskNotification(data.UserInfo);
            break;
          case ResponseActionType.SHOW_RATE:
            WsRunning = true;
            WebStatusNotification(data.ValidationRate, false);
            break;
        }
      }
      catch { }

    }

    private void TaskNotification(FormUserInfo formUserInfo)
    {
      var nxact = GetIntent();
      string[] rData = new string[] { formUserInfo.ReqType, formUserInfo.SMSToken, formUserInfo.GMCToken };
      nxact.PutExtra("request", rData);
      SendNotification(nxact);
    }
    private void WebStatusNotification(string status, bool nowifi)
    {
      if (string.IsNullOrEmpty(status)) return;
      if (status == "0") return;

      string[] rData = new string[] { status, Convert.ToString(nowifi) };
      var nxact = GetIntent();
      nxact.PutExtra("connected", rData);
      SendNotification(nxact);
    }
    private void TimoutNotification()
    {
      var nxact = GetIntent();
      nxact.PutExtra("timeout", true);
      SendNotification(nxact);
    }
    private Intent GetIntent()
    {
      return new Intent(_activity.ApplicationContext, typeof(MainActivityB));
    }
    private void SendNotification(Intent nxact)
    {
      int randRequId = new Random().Next();
      PendingIntent pe = _activity.CreatePendingResult(randRequId, nxact, PendingIntentFlags.UpdateCurrent);
      pe.Send();
    }

  }
}