﻿using NBitcoin;
using EthereumLibrary.Hex.HexConvertors.Extensions;
using System;
using System.Text;
using System.Collections.Generic;
using NoxKeys;

namespace WalletLibrary
{
  public class Device
  {
    BitcoinPubKeyAddress WS_PUB;

    readonly Key TEEKey;

    readonly Guid DeviceID;

    private string WALLETID;

    readonly PubKey TEEPubKey;

    public Device(Key _TEEKey, Guid _DeviceID)
    {
      TEEKey = _TEEKey;
      DeviceID = _DeviceID;
      TEEPubKey = TEEKey.PubKey;
      WALLETID = DeviceID.ToString().ToUpper().Replace("-", "").Substring(0, 6);
      WS_PUB = new BitcoinPubKeyAddress(new KeyId(new byte[] { 146, 17, 90, 20, 206, 41, 255, 170, 171, 24, 82, 234, 70, 248, 115, 63, 247, 55, 93, 61 }), Network.Main);
      MsgCount = 0;
    }

    public Device(string _B58TEEKey, string _DeviceID) : this(new BitcoinSecret(_B58TEEKey).PrivateKey, new Guid(_DeviceID)) { }
    public Device(string _HexTEEKey, Guid _DeviceID) : this(new Key(_HexTEEKey.HexToByteArray(), -1, false), _DeviceID) { }

    public int MsgCount { get; set; }

    public string DevicePubHash() { return TEEKey.GetBitcoinSecret(Network.Main).GetAddress().Hash.ToString(); }
    public string WS_PubHash() { return WS_PUB.Hash.ToString(); }
    public PubKey PubTEE() { return TEEPubKey; }
    public string GetWalletID() { return WALLETID; }
    public string GetDeviceID() { return DeviceID.ToString(); }

    public bool IsErrorKey()
    {
      try
      {
        if (DevicePubHash().ToUpper() == "A7D702365ACF44FE1D2ACE958ACC127FC85DB920")
        {
          return true;
        }
      }
      catch { }
      return false;
    }

    public SignTransferResponse signTransferResponse { get; set; }

    public Dictionary<Guid, NoxShared.SharedSession> sharedSessions = new Dictionary<Guid, NoxShared.SharedSession>();
    public List<byte> Get3DGSaltHash(byte[] _noxSalt)
    {
      using (NoxManagedArray _Resized3DGSalt = new NoxManagedArray(_noxSalt, 3))
      {
        using (var hmac = new System.Security.Cryptography.HMACSHA256(TEEKey.ToBytes()))
        {
          NoxManagedArray khash = new NoxManagedArray(hmac.ComputeHash(_Resized3DGSalt.Value));
          NoxManagedArray shash = new NoxManagedArray(hmac.ComputeHash(DeviceID.ToByteArray()));

          var scrypt = SCrypt.ComputeDerivedKey(khash, shash, 1024, 1, 1, 2, 4).Result;

          return Sclear.EncodeB58NoxHash(scrypt.Value, 0, scrypt.Value.Length);

          khash.Dispose();
          shash.Dispose();
          scrypt.Dispose();

        }
      }
    }

    public string SignMsg(string msg) { return TEEKey.SignMessage(msg); }
    public bool ValidMsg(string msg, string signature)
    {
      string hash = SHashSHA1(msg);
      return WS_PUB.VerifyMessage(hash, signature);
    }
    public string SHashSHA1(string msg)
    {
      byte[] input = System.Text.Encoding.ASCII.GetBytes(msg);

      using (System.Security.Cryptography.SHA1Managed sha1 = new System.Security.Cryptography.SHA1Managed())
      {
        return sha1.ComputeHash(input).ToHex().ToUpper();
      }
    }

  }

}
