﻿using System;
using EthereumLibrary.BouncyCastle.Asn1;
using EthereumLibrary.BouncyCastle.Security;

namespace EthereumLibrary.BouncyCastle.Crypto.Parameters
{
    internal class ECKeyGenerationParameters
        : KeyGenerationParameters
    {
        private ECDomainParameters domainParams;
        private DerObjectIdentifier publicKeyParamSet;

        public ECKeyGenerationParameters(
            ECDomainParameters domainParameters,
            SecureRandom random)
            : base(random, domainParameters.N.BitLength)
        {
            this.domainParams = domainParameters;
        }

        public ECKeyGenerationParameters(
            DerObjectIdentifier publicKeyParamSet,
            SecureRandom random)
            : this(ECKeyParameters.LookupParameters(publicKeyParamSet), random)
        {
            this.publicKeyParamSet = publicKeyParamSet;
        }

        public ECDomainParameters DomainParameters
        {
            get { return domainParams; }
        }

        public DerObjectIdentifier PublicKeyParamSet
        {
            get { return publicKeyParamSet; }
        }
    }
}
