﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Net;
using System.IO;
using System.Security.Cryptography.X509Certificates;
using System.Net.Security;
using NoxKeys;
using System.Threading.Tasks;
using System.Threading;
using WalletLibrary;

namespace NoxService.NWS
{
  public struct HWSResponse
  {
    public bool WebError { get; set; }
    public bool Notfify { get; set; }
    public FormUserInfo formUserInfo { get; set; }
  }
  public class HWS
  {
    NOXWS2 serv;
    public HWS()
    {
      serv = new NOXWS2();
      serv.ConnectionGroupName = NoxService.NXSC.NXSC.ConnectionGroup1;
    }
    public async Task<HWSResponse> GetCurrentSMSToken(Device noxDevice, CancellationToken cancellationToken)
    {
      try
      {
        string SMSTokenMsg = "92115a14ce29ffaaab1852ea46f8733ff7375d3d";
        string msg = noxDevice.SHashSHA1(Guid.NewGuid().ToString());
        string signature = noxDevice.SignMsg(msg + SMSTokenMsg);

        serv.Timeout = 5000;
        var resp = await GetCurrentSMSTokenTask(serv, signature, msg, noxDevice, cancellationToken);

        HWSResponse hWSResponse = new HWSResponse();

        if (resp.Error != null)
        {
          hWSResponse.WebError = true;
        }
        else
        {
          if (!resp.Cancelled)
          {
            if (resp.Result != null && resp.Result.Response != null)
            {
              hWSResponse.formUserInfo = resp.Result.Response;

              if (noxDevice.ValidMsg(hWSResponse.formUserInfo.SMSToken + hWSResponse.formUserInfo.ReqType, resp.Result.Signature))
              {
                hWSResponse.Notfify = true;
              }
            }
          }
        }

        return hWSResponse;

      }
      catch (WebException)
      {

        return new HWSResponse() { WebError = true };
      }
      catch (Exception ex)
      {

        return new HWSResponse() { WebError = true };
      }

    }

    private TaskCompletionSource<GetCurrentSMSTokenCompletedEventArgs> ntask;

    void NOXWSCompleted<T>(TaskCompletionSource<T> tcs, System.ComponentModel.AsyncCompletedEventArgs e, Func<T> getResult)
    {
      try
      {

        if (e.Error != null)
        {
          tcs.TrySetException(e.Error);
        }
        else if (e.Cancelled)
        {
          tcs.TrySetCanceled();
        }
        else
        {

          tcs.TrySetResult(getResult());
        }
      }
      catch (WebException)
      {

        tcs.TrySetCanceled();
      }
      catch
      {
        tcs.TrySetCanceled();
      }
    }
    Task<GetCurrentSMSTokenCompletedEventArgs> GetCurrentSMSTokenTask(NOXWS2 serv, string signature, string msg, Device noxDevice, CancellationToken cancellationToken)
    {
      ntask = new TaskCompletionSource<GetCurrentSMSTokenCompletedEventArgs>();
      serv.GetCurrentSMSTokenCompleted += (x, e) => NOXWSCompleted(ntask, e, () => e);

      cancellationToken.Register(() =>
      {
        ntask.TrySetCanceled();
      });

      serv.GetCurrentSMSTokenAsync(signature, msg, noxDevice.DevicePubHash(), noxDevice.GetDeviceID(), noxDevice.GetWalletID());
      return ntask.Task;
    }

  }
}