﻿using NBitcoin;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Http;
using System.Net.Security;
using System.Security.Cryptography.X509Certificates;
using System.Text;
using System.Threading.Tasks;

namespace NoxService.NXSC
{
  public class NXSC
  {

    public static bool Validator(object sender, X509Certificate cert, X509Chain chain, SslPolicyErrors sslPolicyErrors)
    {

      if (_CertDictionary.ContainsKey(cert.GetPublicKeyString()))
      {
        if (sslPolicyErrors == System.Net.Security.SslPolicyErrors.None)
        {
          return true;
        }
      }

      return false;
    }

    static Dictionary<string, string> _CertDictionary = new Dictionary<string, string>();

    public static string ConnectionGroup1 = Guid.NewGuid().ToString();
    public static string ConnectionGroup2 = Guid.NewGuid().ToString();

    static BitcoinPubKeyAddress PUB = new BitcoinPubKeyAddress(new KeyId(new byte[] { 146, 17, 90, 20, 206, 41, 255, 170, 171, 24, 82, 234, 70, 248, 115, 63, 247, 55, 93, 61 }), Network.Main);

    static bool ValidMsg(string PubKey, string Signature)
    {

      return PUB.VerifyMessage(PubKey, Signature);
    }

    public async static Task<string> GetRateSpot()
    {
      ServicePointManager.ServerCertificateValidationCallback = Validator;
      var captive = await GetCaptiveAsync();

      if (captive != null)
      {

        bool updated = false;

        foreach (NoxDomain noxDomain in captive.noxDomain)
        {
          if (!_CertDictionary.ContainsKey(noxDomain.PubKey) || _CertDictionary[noxDomain.PubKey] != noxDomain.Signature)
          {
            if (ValidMsg(noxDomain.PubKey, noxDomain.Signature))
            {
              _CertDictionary[noxDomain.PubKey] = noxDomain.Signature;
              updated = true;
            }
          }
        }

        if (updated)
        {
          ConnectionGroup1 = Guid.NewGuid().ToString();
          ConnectionGroup2 = Guid.NewGuid().ToString();

        }

        return captive.bTCRate.USD;
      }

      return null;

    }

    async static Task<NoxCaptive> GetCaptiveAsync()
    {
      try
      {
        using (var client = new System.Net.Http.HttpClient())
        {

          client.Timeout = TimeSpan.FromMilliseconds(5000);

          using (var response = await client.GetAsync("http://bitfi.dev/NoxMessages/DMA3Device.aspx?DomainRequest=1"))
          {
            var responseString = await response.Content.ReadAsStringAsync();
            return Newtonsoft.Json.JsonConvert.DeserializeObject<NoxCaptive>(responseString);
          }
        }
      }
      catch (HttpRequestException)
      {
        
      }
      catch(Exception)
      {

      }

      return null;
    }

  }

  public class NoxCaptive
  {

    public NoxDomain[] noxDomain { get; set; }
    public BTCRate bTCRate { get; set; }

  }
  public class NoxDomain
  {

    public string PubKey { get; set; }

    public string Signature { get; set; }

  }
  public class BTCRate
  {
    public string USD { get; set; }
    public string Signature { get; set; }
    public DateTime RateTime { get; set; }
  }
}
